﻿using EFCoreAddRelationship.Entities;
using System;
using System.Linq;
using System.Collections.Generic;
using Microsoft.EntityFrameworkCore;

namespace EFCoreAddRelationship
{
    class Program
    {
        /// <summary>
        /// 初始化Person表和Book表的数据，没有设置Book表的外键列PersonCode的值
        /// </summary>
        static void InitData()
        {
            //初始化数据库数据
            using (var dbContext = new TestDBContext())
            {
                var james = new Person() { PersonCode = "P001", Name = "James", Age = 30 };

                dbContext.Person.Add(james);

                var chineseBook = new Book() { BookCode = "B001", Isbn = "001", BookName = "Chinese" };//没有设置Book表中外键列PersonCode的值
                var japaneseBook = new Book() { BookCode = "B002", Isbn = "001", BookName = "Japanese" };//没有设置Book表中外键列PersonCode的值
                var englishBook = new Book() { BookCode = "B003", Isbn = "001", BookName = "English" };//没有设置Book表中外键列PersonCode的值

                //插入三条数据到Book表
                dbContext.Book.Add(chineseBook);
                dbContext.Book.Add(japaneseBook);
                dbContext.Book.Add(englishBook);

                dbContext.SaveChanges();
            }
        }

        /// <summary>
        /// 删除Person表和Book表的所有数据
        /// </summary>
        static void DeleteAllData()
        {
            using (var dbContext = new TestDBContext())
            {
                dbContext.Database.ExecuteSqlCommand("DELETE FROM [dbo].[Book]");
                dbContext.Database.ExecuteSqlCommand("DELETE FROM [dbo].[Person]");
            }
        }

        /// <summary>
        /// 不正确地设置Person表和Book表的关联关系，这种方法会让EF Core错误地生成INSERT语句，而不是UPDATE语句
        /// </summary>
        static void SetRelationshipIncorrectly()
        {
            using (var dbContext = new TestDBContext())
            {
                var james = dbContext.Person.First(e => e.Name == "James");//首先通过DbContext从数据库中查询出要建立关联关系的Person表实体

                var chineseBook = new Book() { BookCode = "B001" };//只构造Book实体的Key属性即可，根据BookCode值"B001"来构造Chinese Book
                var japaneseBook = new Book() { BookCode = "B002" };//只构造Book实体的Key属性即可，根据BookCode值"B002"来构造Japanese Book
                var englishBook = new Book() { BookCode = "B003" };//只构造Book实体的Key属性即可，根据BookCode值"B003"来构造English Book

                Console.WriteLine($"Before adding, chineseBook entity state is :{dbContext.Entry(chineseBook).State.ToString()}");//可以看到由于此时Book实体chineseBook没有被DbContext跟踪，所以状态是Detached
                Console.WriteLine($"Before adding, japaneseBook entity state is :{dbContext.Entry(japaneseBook).State.ToString()}");//可以看到由于此时Book实体japaneseBook没有被DbContext跟踪，所以状态是Detached
                Console.WriteLine($"Before adding, englishBook entity state is :{dbContext.Entry(englishBook).State.ToString()}");//可以看到由于此时Book实体englishBook没有被DbContext跟踪，所以状态是Detached

                Console.WriteLine();

                james.Book = new List<Book>();//由于我们在上面调用dbContext.Person.First(e => e.Name == "James")时，没有用EF Core中Eager Loading的Include方法来加载Book实体集合，所以这里要用List类来构造一个Book实体集合，否则james.Book为null

                james.Book.Add(chineseBook);//添加chineseBook到Person类的Book实体集合
                Console.WriteLine("chineseBook was added into Person.Book collection");

                james.Book.Add(japaneseBook);//添加japaneseBook到Person类的Book实体集合
                Console.WriteLine("japaneseBook was added into Person.Book collection");

                james.Book.Add(englishBook);//添加englishBook到Person类的Book实体集合
                Console.WriteLine("englishBook was added into Person.Book collection");

                Console.WriteLine();

                Console.WriteLine($"After querying DbContext.Entry(chineseBook), chineseBook entity state is :{dbContext.Entry(chineseBook).State.ToString()}");//调用DbContext.Entry()方法后，DbContext发现一个原本状态是Detached的Book实体chineseBook被加入到Person.Book集合中了，所以此时chineseBook的实体状态变为了Added
                Console.WriteLine($"After querying DbContext.Entry(japaneseBook), japaneseBook entity state is :{dbContext.Entry(japaneseBook).State.ToString()}");//调用DbContext.Entry()方法后，DbContext发现一个原本状态是Detached的Book实体japaneseBook被加入到Person.Book集合中了，所以此时japaneseBook的实体状态变为了Added
                Console.WriteLine($"After querying DbContext.Entry(englishBook), englishBook entity state is :{dbContext.Entry(englishBook).State.ToString()}");//调用DbContext.Entry()方法后，DbContext发现一个原本状态是Detached的Book实体englishBook被加入到Person.Book集合中了，所以此时englishBook的实体状态变为了Added

                dbContext.SaveChanges();//由于此时chineseBook、japaneseBook和englishBook的EntityState都是Added，所以此时DbContext.SaveChanges方法调用后，EF Core生成的是INSERT语句，将chineseBook、japaneseBook和englishBook插入数据库表Book，导致插入了重复值到唯一键列BookCode，所以数据库报错
            }
        }

        /// <summary>
        /// 正确地设置Person表和Book表的关联关系，这种方法会让EF Core正确地生成UPDATE语句，在数据库中设置Book表的PersonCode列数据
        /// </summary>
        static void SetRelationshipCorrectly()
        {
            using (var dbContext = new TestDBContext())
            {
                var james = dbContext.Person.First(e => e.Name == "James");//首先通过DbContext从数据库中查询出要建立关联关系的Person表实体

                var chineseBook = new Book() { BookCode = "B001" };//只构造Book实体的Key属性即可，根据BookCode值"B001"来构造Chinese Book
                var japaneseBook = new Book() { BookCode = "B002" };//只构造Book实体的Key属性即可，根据BookCode值"B002"来构造Japanese Book
                var englishBook = new Book() { BookCode = "B003" };//只构造Book实体的Key属性即可，根据BookCode值"B003"来构造English Book

                dbContext.Attach(chineseBook);//将chineseBook关联到DbContext，开始跟踪
                dbContext.Attach(japaneseBook);//将japaneseBook关联到DbContext，开始跟踪
                dbContext.Attach(englishBook);//将englishBook关联到DbContext，开始跟踪

                Console.WriteLine($"After querying DbContext.Entry(chineseBook), chineseBook entity state is :{dbContext.Entry(chineseBook).State.ToString()}");//由于上面chineseBook被Attach到DbContext开始跟踪了，所以此时chineseBook的实体状态是Unchanged
                Console.WriteLine($"After querying DbContext.Entry(japaneseBook), japaneseBook entity state is :{dbContext.Entry(japaneseBook).State.ToString()}");//由于上面japaneseBook被Attach到DbContext开始跟踪了，所以此时japaneseBook的实体状态是Unchanged
                Console.WriteLine($"After querying DbContext.Entry(englishBook), englishBook entity state is :{dbContext.Entry(englishBook).State.ToString()}");//由于上面englishBook被Attach到DbContext开始跟踪了，所以此时englishBook的实体状态是Unchanged

                Console.WriteLine();

                james.Book = new List<Book>();//由于我们在上面调用dbContext.Person.First(e => e.Name == "James")时，没有用EF Core中Eager Loading的Include方法来加载Book实体集合，所以这里要用List类来构造一个Book实体集合，否则james.Book为null

                james.Book.Add(chineseBook);//添加chineseBook到Person类的Book实体集合
                Console.WriteLine("chineseBook was added into Person.Book collection");

                james.Book.Add(japaneseBook);//添加japaneseBook到Person类的Book实体集合
                Console.WriteLine("japaneseBook was added into Person.Book collection");

                james.Book.Add(englishBook);//添加englishBook到Person类的Book实体集合
                Console.WriteLine("englishBook was added into Person.Book collection");

                Console.WriteLine();

                Console.WriteLine($"Berfore querying DbContext.Entry(chineseBook), chineseBook.PersonCode is :{chineseBook.PersonCode ?? "null"}");//此时由于我们还没有调用DbContext.Entry()方法，所以DbContext还无法察觉到chineseBook已经被添加到Person类的Book实体集合了，所以chineseBook.PersonCode为null
                Console.WriteLine($"After querying DbContext.Entry(chineseBook), chineseBook entity state is :{dbContext.Entry(chineseBook).State.ToString()}");//调用DbContext.Entry()方法后，DbContext发现一个原本状态是Unchanged的Book实体chineseBook被加入到Person.Book集合中了，所以此时chineseBook的实体状态变为了Modified
                Console.WriteLine($"After querying DbContext.Entry(chineseBook),  chineseBook.PersonCode is :{chineseBook.PersonCode}");//由于上面我们调用DbContext.Entry(chineseBook)使得DbContext得知了chineseBook被加入到Person.Book集合中了，所以DbContext还将Book实体的外键属性PersonCode也进行了赋值，为P001

                Console.WriteLine();

                Console.WriteLine($"Berfore querying DbContext.Entry(japaneseBook),japaneseBook.PersonCode is :{japaneseBook.PersonCode ?? "null"}");//很有意思的是我们上面在chineseBook上调用DbContext.Entry()方法后，japaneseBook的PersonCode属性也不为null了，变为了P001，说明调用一次DbContext.Entry()方法后，会引发DbContext重新检查所有被跟踪实体的状态
                Console.WriteLine($"After querying DbContext.Entry(japaneseBook), japaneseBook entity state is :{dbContext.Entry(japaneseBook).State.ToString()}");//在上面为chineseBook调用DbContext.Entry()方法时，DbContext同时发现了原本状态是Unchanged的Book实体japaneseBook，也被加入到了Person.Book集合中，所以japaneseBook的实体状态也变为了Modified
                Console.WriteLine($"After querying DbContext.Entry(japaneseBook),  japaneseBook.PersonCode is :{japaneseBook.PersonCode}");//在上面为chineseBook调用DbContext.Entry()方法时,DbContext得知了japaneseBook也被加入到了Person.Book集合中，所以DbContext将japaneseBook的PersonCode属性也赋值为P001了

                Console.WriteLine();

                Console.WriteLine($"Berfore querying DbContext.Entry(englishBook),englishBook.PersonCode is :{englishBook.PersonCode ?? "null"}");//在上面为chineseBook调用DbContext.Entry()方法时,DbContext得知了englishBook也被加入到了Person.Book集合中，所以DbContext将englishBook的PersonCode属性也赋值为P001了
                Console.WriteLine($"After querying DbContext.Entry(englishBook), englishBook entity state is :{dbContext.Entry(englishBook).State.ToString()}");//在上面为chineseBook调用DbContext.Entry()方法时，DbContext同时发现了原本状态是Unchanged的Book实体englishBook，也被加入到了Person.Book集合中，所以englishBook的实体状态也变为了Modified
                Console.WriteLine($"After querying DbContext.Entry(englishBook),  englishBook.PersonCode is :{englishBook.PersonCode}");//在上面为chineseBook调用DbContext.Entry()方法时,DbContext得知了englishBook也被加入到了Person.Book集合中，所以DbContext将englishBook的PersonCode属性也赋值为P001了

                dbContext.SaveChanges();//由于此时chineseBook、japaneseBook和englishBook的EntityState都是Modified，所以此时DbContext.SaveChanges方法调用后，EF Core生成的是UPDATE语句，通过更新数据库Book表的PersonCode列，将Chinese、Japanese和English三行Book数据同Person表的数据成功关联了起来
            }
        }


        static void Main(string[] args)
        {
            DeleteAllData();//调用DeleteAllData方法删除Person表和Book表的所有数据，防止有脏数据
            InitData();//初始化Person表和Book表的数据

            SetRelationshipCorrectly();//正确地的设置Person表和Book表的关联关系
            //SetRelationshipIncorrectly();//不正确地的设置Person表和Book表的关联关系，该方法会抛出异常错误

            Console.WriteLine("Press any key to quit...");
            Console.ReadKey();
        }
    }
}
